module UU_Scanner where

{- A parametrisable scanner
 -
 - Author(s): Doaitse Swierstra
 - Version 1.0, May 25, 1998: first appearance on the software web site.
 -}
import Char
import UU_Parsing
import UU_Parsing_ext

data Assoc = LeftAssoc | RightAssoc | NonAssoc | BothAssoc deriving Eq

data TokenType 
  = TkSymbol
  | TkOperscan
  | TkOperget
  | TkKeyword   
  | TkString
  | TkInteger  
  | TkVarid 
  | TkConid  
  | TkError 
  deriving Eq

type Linenumber = Int
data Token = Tok TokenType String Linenumber

instance Eq Token where
  Tok TkSymbol              x _ == Tok TkSymbol                y _ = x == y
  Tok TkKeyword             x _ == Tok TkKeyword               y _ = x == y
  Tok TkOperscan            x _ == Tok TkOperscan              y _ = x == y
  Tok TkOperscan            _ _ == Tok TkOperget               _ _ = True
  Tok a                     _ _ == Tok b                       _ _ = a == b  
           
               
maybeshow 0 = ""
maybeshow n = " at line " ++ show n

instance Show Token where
  showsPrec _ token
    = showString 
       (case token of
        Tok TkSymbol          s i -> "Symbol   " ++ s ++  maybeshow i
        Tok TkOperscan        s i -> "Operator " ++ s ++  maybeshow i
        Tok TkOperget         s i -> "Op      "  ++ s ++  maybeshow i
        Tok TkKeyword         s i -> "Keyword "  ++ s ++  maybeshow i
        Tok TkString          s i -> "String \"" ++ s ++ "\""
        Tok TkInteger         s i -> "Integer "  ++ s ++  maybeshow i
        Tok TkVarid           s i -> "Lower case identifier "++ s ++  maybeshow i
        Tok TkConid           s i -> "Upper case identifier "++ s ++  maybeshow i
        Tok TkError           s i -> "error in scanner: "    ++ s ++  maybeshow i
        )

isspace c     =  c == ' '   || c == '\t'  || c == '\r'  ||
                               c == '\f'  || c == '\v'
                               

scanner keywordstxt keywordsops specchars opchars
  = scan
    where scan :: String -> Linenumber ->[Token]
          scan [] lc      = []
          scan (c:s) lc
               | isspace c = scan (dropWhile isspace s) lc  
          scan ('\n':s) lc = scan s (lc + 1)
          scan ('-':'-':'>':s) lc
            = let (line, rest) = span (/= '\n') s
              in line `seq` ((Tok TkString line lc):scan rest lc )         
          scan ('-':'-':s) lc
            = scan (dropWhile (/= '\n') s) lc          
          scan ('{':'-':s) lc
            = lexNest scan s lc
            where
              lexNest f ('\n':s)    lc = lexNest f s (lc+1)
              lexNest f ('-':'}':s) lc = f s lc
              lexNest f ('{':'-':s) lc = lexNest (lexNest f) s lc
              lexNest f (c:s)       lc = lexNest f s lc
              lexNest _ ""          lc
                = [ Tok TkError "Unterminated nested comment" lc]          
          scan ('"':s) lc
            | head zs /= '"' 
                = Tok TkError ("Unterminated string") lc : scan (tail zs) lc
            | otherwise 
                = Tok TkString ys lc                     : scan (tail zs) lc
            where
              (ys, zs) = span (not . isEnd) s
              isEnd c = c == '"' || c == '\n'           
          scan (c:s) lc
            | isSymbol c 
                = Tok TkSymbol [c] lc : scan s lc
            | isLower c || isUpper c
                = let (name, s') = span (\ c -> isAlphaNum c || (c =='_') || (c == '\'')) (c:s)
                  in if name `elem` keywordstxt 
                      then Tok TkKeyword name lc: scan s' lc
                      else Tok (if isLower c then TkVarid else TkConid) name lc: scan s' lc
            | isOpsym c
                = let (name, s') = span isOpsym (c:s)
                  in if name `elem` keywordsops 
                      then Tok TkKeyword  name lc: scan s' lc
                      else Tok TkOperscan name lc: scan s' lc
            | isDigit c = let (number,s') = span isDigit (c:s)
                          in  (Tok TkInteger number lc):scan s' lc 
            | otherwise
                = (Tok TkError ("Unexpected character '"++[c]++"'") lc):scan s lc
            where
              isOpsym c  = c `elem` opchars
              isSymbol c = c `elem` specchars

-------------------------------------------------------------------------
-- Parsers for  Symbols 
-------------------------------------------------------------------------
instance Symbol Token

get_tok_val (Tok _ s _) = s
get_tok_line (Tok _ _ n) = n

pOper_any
  =  get_tok_val <$> symbol (Tok TkOperget "?OPER?" 0)
  
pOper name 
  =  get_tok_val <$> symbol (Tok TkOperscan name                 0)

pOperLine name 
  =  get_tok_line <$> symbol (Tok TkOperscan name                 0)

pKey keyword
  =  get_tok_val <$> symbol (Tok TkKeyword keyword               0)

pSym s
  =  get_tok_val <$> symbol (Tok TkSymbol [s]                    0)

pSymLine s
  =  get_tok_line <$> symbol (Tok TkSymbol [s]                    0)

pString
  =  get_tok_val <$> symbol (Tok TkString "?STR?"      0)
 
pInteger
  =  get_tok_val <$> symbol (Tok TkInteger "1"                   0)

pVarid 
  =  get_tok_val <$> symbol (Tok TkVarid  "?LC?" 0)

pConid 
  =  get_tok_val <$> symbol (Tok TkConid  "?UC?" 0)


pcommas p = pListSep (pSym ',') p 
psemics p = pListSep (pSym ';') p  
pparens p = pSym  '(' <-*>  p <*-> pSym ')' 
pbracks p = pSym  '[' <-*>  p <*-> pSym ']' 
pcurly  p = pSym  '{' <-*>  p <*-> pSym '}' 

pparens_pcommas :: Parser Token a -> Parser Token [a]
pbracks_pcommas :: Parser Token a -> Parser Token [a]
pcurly_psemics  :: Parser Token a -> Parser Token [a]
pparens_pcommas = pparens.pcommas
pbracks_pcommas = pbracks.pcommas
pcurly_psemics  = pcurly .psemics

pcurly_psemics_lines p = (\ a b x -> (b,a):x) <$> pSymLine '{' <*> p <*> tail
                         where tail = ((\ a b x -> (b,a):x) <$> pSymLine ';' <*> p <*> tail) <|>
						              (const [] <$> pSym '}')